// No copyright, no warranty; use as you will.
// Written by Ronald Bourret, Technical University of Darmstadt, 1998-9

import de.tudarmstadt.ito.schemas.dtd.DTD;
import de.tudarmstadt.ito.schemas.converters.DDMLToDTD;
import de.tudarmstadt.ito.schemas.converters.DTDToDDML;
import de.tudarmstadt.ito.schemas.converters.SubsetToDTD;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.Hashtable;
import org.xml.sax.InputSource;
import org.xml.sax.Parser;

// Imports for the Oracle version 2 parser
import oracle.xml.parser.v2.SAXParser;

/*
// Imports for the Sun parser (none needed -- see getSAXParser())
*/

/*
// Imports for the Xerces parser
import org.apache.xerces.parsers.SAXParser;
*/

/**
 * This application converts a schema document to a DTD and vice versa.
 * Currently, it only supports conversions from DDML documents to DTDs and
 * vice versa. To use other schema languages, you need to write the appropriate
 * converters. (It should take a day or less for each direction assuming you
 * know the schema language well and don't support the more exotic stuff such
 * as schema reuse features.)
 *
 * <P>The command line syntax for this application is:</P>
 *
 * <PRE>
 *    java GenerateMap &lt;input-file&gt;
 * </PRE>
 *
 * <p>This application is also hard-coded to use the Oracle version 2 XML parser.
 * However, it also includes commented-out code to use the Xerces and Sun XML
 * parsers. To modify ConvertSchema for your XML parser, comment/uncomment the
 * correct import statements and version of getSAXParser(). If you are using a
 * parser other than Oracle version 2, Xerces, or Sun, you will
 * need to modify the code accordingly.</p>
 */

public class ConvertSchema
{
   // ***********************************************************************
   // Main methods
   // ***********************************************************************

   public static void main (String[] argv)
   {
      try
      {
         if (argv.length != 1)
         {
            System.out.println("\nUsage: java ConvertSchema <schema-file>\n");
            return;
         }

         convert(argv[0]);
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   static void convert(String filename)
      throws Exception
   {
      InputSource src;
      String      extension, basename;

      // Create an InputSource over the file.

      src = new InputSource(getFileURL(filename));

      // Get the base name and file extension.

      extension = getExtension(filename);
      basename = getBasename(filename);

      // Determine the document type from its extension: .DDM for DDML and .DTD
      // for external subsets.

      if (extension.equals("DDM"))
      {
         // DDML Document
         convertDDMLToDTD(src, basename);
      }
      else if (extension.equals("DTD"))
      {
         // External subset (DTD)
         convertDTDToDDML(src, basename);
      }
      else
      {
         throw new Exception("Unrecognized file extension: " + extension);
      }
   }

   static void convertDDMLToDTD(InputSource src, String basename)
      throws Exception
   {
      DDMLToDTD        ddmlToDTD;
      DTD              dtd;
      FileOutputStream dtdFile;

      // Create the DTD file.

      dtdFile = new FileOutputStream(basename + ".dtd");

      // Convert the DDML document to a DTD object, then serialize
      // the DTD object to the DTD file.

      ddmlToDTD = new DDMLToDTD(getSAXParser());
      dtd = ddmlToDTD.convert(src);
      dtd.serialize(dtdFile, true);

      // Close the DTD file.

      dtdFile.close();
   }

   static void convertDTDToDDML(InputSource src, String basename)
      throws Exception
   {
      SubsetToDTD      subsetToDTD;
      DTDToDDML        dtdToDDML;
      DTD              dtd;
      FileOutputStream ddmlFile;
      Hashtable        namespaceURIs = null;

      // Create the DDML file.

      ddmlFile = new FileOutputStream(basename + ".ddm");

      // Convert the DTD to a DTD object, then serialize the DTD object
      // as a DDML document.

      // If there are namespace prefixes in your DTD, you need to declare
      // them here, putting each one in the namespaceURIs hashtable with
      // the prefix as the key and the namespace URI as the value. For
      // example, the following commented-out lines state that the DDML
      // prefix corresponds to the http://www.purl.org/NET/ddml/v1 URI.
      // Note that the default namespace can be declared by adding an
      // empty string ("") as a key.

//      namespaceURIs = new Hashtable();
//      namespaceURIs.put("DDML", "http://www.purl.org/NET/ddml/v1");

      subsetToDTD = new SubsetToDTD();
      dtd = subsetToDTD.convertExternalSubset(src, namespaceURIs);
      dtdToDDML = new DTDToDDML();
      dtdToDDML.convert(dtd, ddmlFile, true, 3);

      // Close the DDML file.

      ddmlFile.close();
   }

   // ***********************************************************************
   // General utility methods
   // ***********************************************************************

   static String getFileURL(String fileName)
   {
      File   file;

      // Construct a file URL for the file.

      file = new File(fileName);
      return "file:///" + file.getAbsolutePath();
   }

   static String getBasename(String filename)
   {
      int    period;

      // Get the basename of the file.

      period = filename.lastIndexOf('.', filename.length());
      if (period == -1)
      {
         return filename;
      }
      else
      {
         return filename.substring(0, period);
      }
   }

   static String getExtension(String filename)
   {
      int    period;

      // Get the file extension.

      period = filename.lastIndexOf('.', filename.length());
      if (period == -1)
      {
         return "";
      }
      else
      {
         return filename.substring(period + 1, filename.length()).toUpperCase();
      }
   }

   // ***********************************************************************
   // Methods that use the Oracle version 2 parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************
   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Oracle parser.

      SAXParser parser;

      // Instantiate the parser and set various options
      parser = new SAXParser();
      parser.setValidationMode(true);
      return parser;
   }

   // ***********************************************************************
   // Methods that use the Sun parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************

/*
   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Sun parser.

      return new com.sun.xml.parser.Parser();
   }
*/

   // ***********************************************************************
   // Methods that use the Xerces parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************

/*
   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Xerces parser.

      return new SAXParser();
   }
*/
}

