// No copyright, no warranty; use as you will.
// Written by Ronald Bourret, Technical University of Darmstadt, 1998-9

import de.tudarmstadt.ito.xmldbms.DBMSToDOM;
import de.tudarmstadt.ito.xmldbms.DOMToDBMS;
import de.tudarmstadt.ito.xmldbms.Map;
import de.tudarmstadt.ito.xmldbms.helpers.KeyGeneratorImpl;
import de.tudarmstadt.ito.xmldbms.mapfactories.MapFactory_MapDocument;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import org.xml.sax.InputSource;
import org.xml.sax.Parser;
import org.w3c.dom.Document;

// Imports for the Oracle version 2 parser
import de.tudarmstadt.ito.domutils.DF_Oracle2;
import de.tudarmstadt.ito.domutils.NQ_Oracle2;
import oracle.xml.parser.v2.DOMParser;
import oracle.xml.parser.v2.SAXParser;
import oracle.xml.parser.v2.XMLDocument;

/*
// Imports for the Sun parser
import de.tudarmstadt.ito.domutils.DF_Sun;
import de.tudarmstadt.ito.domutils.NQ_Sun;
import com.sun.xml.tree.XmlDocument;
import com.sun.xml.tree.XmlDocumentBuilder;
*/

/*
// Imports for the Xerces parser
import de.tudarmstadt.ito.domutils.DF_Xerces;
import de.tudarmstadt.ito.domutils.NQ_DOM2;
import org.apache.xerces.parsers.DOMParser;
import org.apache.xerces.parsers.SAXParser;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
*/

/**
 * This application accepts a map document name, an XML document name, and
 * (in the case of transferring data from the database to an XML document) a
 * table name and key value. It transfers data in the specified direction.
 *
 * <P>The command line syntax for this application is:</P>
 *
 * <PRE>
 *    java Transfer {-todbms|-toxml} &lt;map-file&gt; &lt;xml-file&gt; [&lt;table-name&gt; &lt;key-value&gt;...]
 * </PRE>
 *
 * <P>where:</P>
 *
 * <pre>
 *    -todbms | -toxml indicates the direction of data transfer
 *    &lt;map-file&gt; is the name of the XML-DBMS map document
 *    &lt;xml-file&gt; is the name of the input or output XML document
 *    &lt;table-name&gt; is the name of the table containing the data to retrieve
 *       (-toxml direction only)
 *    &lt;key-value&gt;... is one or more values in a (multi-part) key
 *       (-toxml direction only)
 * </pre>
 *
 * <P>For this application to work, there must be an ODBC data source named
 * "xmldbms" and the tables used to store data must have already been created
 * in the data source. This application uses the Sun JDBC-ODBC Bridge, but
 * includes commented-out code for using the Easysoft JDBC-ODBC Bridge. If
 * you are using a different JDBC driver, you will need to modify the code
 * to use your driver.</p>
 *
 * <p>This application is also hard-coded to use the Oracle version 2 XML parser
 * and DOM implementation. However, it also includes commented-out code to use
 * the Xerces and Sun XML parsers and DOM implementation. To modify Transfer
 * for your XML parser and DOM implementation, comment/uncomment the appropriate
 * import statements, code in toDBMS() and toXML(), versions of
 * getSAXParser(), openDocument(), and writeDocument(). If you are using a parser
 * or DOM implementation other than Oracle version 2, Xerces, or Sun, you will
 * need to modify the code accordingly.</p>
 *
 * <p>If the map document specifies that keys are to be generated, the default
 * KeyGenerator implementation is used; for more information, see
 * de.tudarmstadt.ito.xmldbms.helper.KeyGeneratorImpl.</P>
 */

public class Transfer
{
   // ***********************************************************************
   // Main methods
   // ***********************************************************************

   public static void main (String[] argv)
   {
      String   mapFilename = null,
               xmlFilename = null,
               url = "jdbc:odbc:xmldbms",
//               url = "jdbc:easysoft://localhost:8831/xmldbms",
               tableName = null;
      Object[] key = null;
      boolean  toxml;

      try
      {
         if (argv.length < 3) throw new IllegalArgumentException();

         if ((!argv[0].equals("-todbms")) && (!argv[0].equals("-toxml")))
            throw new IllegalArgumentException();

         toxml = (argv[0].equals("-toxml"));
         mapFilename = argv[1];
         xmlFilename = argv[2];
         if (toxml)
         {
            if (argv.length < 5) throw new IllegalArgumentException();

            tableName = argv[3];

            key = new Object[argv.length - 4];
            for (int i = 4; i < argv.length; i++)
            {
               key[i - 4] = argv[i];
            }
         }

         if (toxml)
         {
            toXML(mapFilename, xmlFilename, url, tableName, key);
         }
         else
         {
            toDBMS(mapFilename, xmlFilename, url);
         }
      }
      catch (IllegalArgumentException iae)
      {
         System.out.println("\nUsage: java Transfer {-todbms|-toxml} <map-file> <xml-file> [<table-name> <key-value>...]\n<table-name> and <key-value> are required when -toxml is chosen.\n");
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   static void toDBMS(String mapFilename, String xmlFilename, String url)
      throws Exception
   {
      Connection       conn1 = null, conn2 = null;
      Map              map;
      Document         doc;
      DOMToDBMS        domToDBMS;
      KeyGeneratorImpl keyGenerator = null;

      try
      {
         // Get a JDBC driver. You will need to modify this code if
         // you are using a different driver.
         Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
//         Class.forName("easysoft.sql.jobDriver");

         // Connect to the database. 
         conn1 = DriverManager.getConnection(url);
         conn2 = DriverManager.getConnection(url);

         // Create and initialize a key generator
         keyGenerator = new KeyGeneratorImpl(conn1);
         keyGenerator.initialize();

         // Create the Map object and open the XML document.
         map = createMap(mapFilename, conn2);
         doc = openDocument(xmlFilename);

         // Create a new DOMToDBMS object and transfer the data. Note that
         // different DOM implementations need different NameQualifier
         // implementations.
         domToDBMS = new DOMToDBMS(map, keyGenerator, new NQ_Oracle2());
//         domToDBMS = new DOMToDBMS(map, keyGenerator, new NQ_Sun());
//         domToDBMS = new DOMToDBMS(map, keyGenerator, new NQ_DOM2()); // Xerces
         domToDBMS.storeDocument(doc);
      }
      finally
      {
         if (conn1 != null) conn1.close();
         if (conn2 != null) conn2.close();
      }
   }

   static void toXML(String mapFilename, String xmlFilename, String url, String tableName, Object[]key)
      throws Exception
   {
      Connection       conn = null;
      Map              map;
      DBMSToDOM        dbmsToDOM;
      Document         doc;

      try
      {
         // Get a JDBC driver. You will need to modify this code if
         // you are using a different driver.
         Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
//         Class.forName("easysoft.sql.jobDriver");

         // Connect to the database
         conn = DriverManager.getConnection(url);

         // Create the Map object.
         map = createMap(mapFilename, conn);

         // Create a new DBMSToDOM object and transfer the data. Note that
         // different DOM implementations need different DocumentFactory
         // implementations.
         dbmsToDOM = new DBMSToDOM(map, new DF_Oracle2());
//         dbmsToDOM = new DBMSToDOM(map, new DF_Sun());
//         dbmsToDOM = new DBMSToDOM(map, new DF_Xerces());
         doc = dbmsToDOM.retrieveDocument(tableName, key);
      }
      finally
      {
         if (conn != null) conn.close();
      }

      // Write the DOM tree to a file.
      writeDocument(doc, xmlFilename);
   }

   // ***********************************************************************
   // General utility methods
   // ***********************************************************************

   static Map createMap(String mapFilename, Connection conn) throws Exception
   {
      MapFactory_MapDocument    factory;

      // Create a new map factory and create the Map.
      factory = new MapFactory_MapDocument(conn, getSAXParser());
      return factory.createMap(new InputSource(getFileURL(mapFilename)));
   }

   static String getFileURL(String fileName)
   {
      File   file;

      file = new File(fileName);
      return "file:///" + file.getAbsolutePath();
   }

   // ***********************************************************************
   // Methods that use the Oracle version 2 parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************

   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Oracle parser.

      SAXParser parser;

      // Instantiate the parser and set various options
      parser = new SAXParser();
      parser.setValidationMode(true);
      return parser;
   }

   static Document openDocument(String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Oracle parser.

      DOMParser parser;

      // Instantiate the parser and set various options.
      parser = new DOMParser();
      parser.setValidationMode(false);
      parser.showWarnings(true);

      // Parse the input file
      parser.parse(new InputSource(getFileURL(xmlFilename)));

      // Return the DOM tree
      return parser.getDocument();
   }

   static void writeDocument(Document doc, String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Oracle parser.

      FileOutputStream xmlFile;

      // Write the DOM tree to a file.
      xmlFile = new FileOutputStream(xmlFilename);
      ((XMLDocument)doc).print((OutputStream)xmlFile);
      xmlFile.close();
   }

   // ***********************************************************************
   // Methods that use the Sun parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************

/*
   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Sun parser.

      return new com.sun.xml.parser.Parser();
   }

   static Document openDocument(String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Sun parser.

      com.sun.xml.parser.Parser parser;
      XmlDocumentBuilder        docBuilder;

      // Instantiate the parser and set various options.
      parser = new com.sun.xml.parser.Parser();
      docBuilder =  new XmlDocumentBuilder();
      parser.setDocumentHandler(docBuilder);

      // Parse the input file
      parser.parse(new InputSource(getFileURL(xmlFilename)));

      // Return the DOM tree
      return docBuilder.getDocument();
   }

   static void writeDocument(Document doc, String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Sun parser.

      FileOutputStream xmlFile;

      // Write the DOM tree to a file.
      xmlFile = new FileOutputStream(xmlFilename);
      ((XmlDocument)doc).write((OutputStream)xmlFile);
      xmlFile.close();
   }
*/

   // ***********************************************************************
   // Methods that use the Xerces parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************

/*
   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Xerces parser.

      return new SAXParser();
   }

   static Document openDocument(String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Xerces parser.

      DOMParser parser;

      // Instantiate the parser and set various options.
      parser = new DOMParser();
      parser.setFeature("http://xml.org/sax/features/namespaces", true);

      // Parse the input file
      parser.parse(new InputSource(getFileURL(xmlFilename)));

      // Return the DOM tree
      return parser.getDocument();
   }

   static void writeDocument(Document doc, String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Xerces parser.

      FileOutputStream xmlFile;
      OutputFormat     format;
      XMLSerializer    serial;

      // Write the DOM tree to a file.
      xmlFile = new FileOutputStream(xmlFilename);
      format = new OutputFormat(doc);
      format.setIndenting(true);
      serial = new XMLSerializer((OutputStream)xmlFile, format);
      serial.asDOMSerializer().serialize(doc);
      xmlFile.close();
   }
*/
}

