// No copyright, no warranty; use as you will.
// Written by Ronald Bourret, Technical University of Darmstadt, 2000

import de.tudarmstadt.ito.xmldbms.DBMSToDOM;
import de.tudarmstadt.ito.xmldbms.Map;
import de.tudarmstadt.ito.xmldbms.mapfactories.MapFactory_MapDocument;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.Statement;
import org.xml.sax.InputSource;
import org.xml.sax.Parser;
import org.w3c.dom.Document;

// Imports for the Oracle version 2 parser
import de.tudarmstadt.ito.domutils.DF_Oracle2;
import oracle.xml.parser.v2.SAXParser;
import oracle.xml.parser.v2.XMLDocument;

/*
// Imports for the Sun parser
import de.tudarmstadt.ito.domutils.DF_Sun;
import com.sun.xml.tree.XmlDocument;
*/

/*
// Imports for the Xerces parser
import de.tudarmstadt.ito.domutils.DF_Xerces;
import org.apache.xerces.parsers.SAXParser;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
*/

/**
 * This application accepts a table name or a SELECT statement, a map document
 * name, and an XML document name. It transfers data from the data or SELECT
 * statement to the XML document according to the map.
 *
 * <P>The command line syntax for this application is:</P>
 *
 * <PRE>
 *    java TransferResultSet {-t &lt;table-name&gt; | -s &lt;SELECT-statement&gt;} &lt;map-file&gt; &lt;xml-file&gt;
 * </PRE>
 *
 * <p>If a SELECT statement is used, it must be enclosed in quotes.</p>
 *
 * <P>For this application to work, there must be an ODBC data source named
 * "xmldbms" and the tables used to store data must have already been created
 * in the data source. This application uses the Sun JDBC-ODBC Bridge, but
 * includes commented-out code for using the Easysoft JDBC-ODBC Bridge. If
 * you are using a different JDBC driver, you will need to modify the code
 * to use your driver.</p>
 *
 * <p>This application is also hard-coded to use the Oracle version 2 XML parser
 * and DOM implementation. However, it also includes commented-out code to use
 * the Xerces and Sun XML parsers and DOM implementation. To modify TransferResultSet
 * for your XML parser and DOM implementation, comment/uncomment the appropriate
 * import statements, code in and toXML(), versions of
 * getSAXParser() and writeDocument(). If you are using a parser
 * or DOM implementation other than Oracle version 2, Xerces, or Sun, you will
 * need to modify the code accordingly.</p>
 */

public class TransferResultSet
{
   public static void main (String[] argv)
   {
      String   mapFilename = null,
               xmlFilename = null,
               url = "jdbc:odbc:xmldbms",
//               url = "jdbc:easysoft://localhost:8831/xmldbms",
               select = null;
      boolean  table;

      try
      {
         if (argv.length != 4) throw new IllegalArgumentException();

         if ((!argv[0].equals("-t")) && (!argv[0].equals("-s")))
            throw new IllegalArgumentException();

         table = (argv[0].equals("-t"));

         // Set the SELECT statement. The commented out SELECT statement is
         // designed for use with the sales_rs.map. Note that it uses MS
         // Access-specific syntax.

         select = argv[1];
         //select = "SELECT Sales.CustNumber, Lines.Part, Sales.Number, Lines.SONumber, Sales.Date, Lines.Number AS LineNumber FROM Sales INNER JOIN Lines ON Sales.Number = Lines.SONumber";

         mapFilename = argv[2];
         xmlFilename = argv[3];

         if (table)
         {
            select = "SELECT * FROM " + select;
         }

         toXML(select, mapFilename, xmlFilename, url);
      }
      catch (IllegalArgumentException iae)
      {
         System.out.println("\nUsage: java TransferResultSet {-t <table-name> | -s <SELECT-statement>} <map-file> <xml-file>\n\nIf a SELECT statement is used, it must be enclosed in quotes.\n");
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   static void toXML(String select, String mapFilename, String xmlFilename, String url)
      throws Exception
   {
      Connection       conn = null;
      Statement        stmt = null;
      ResultSet        rs = null;
      Map              map;
      DBMSToDOM        dbmsToDOM;
      Document         doc;

      try
      {
         // Connect to the database and get the result set
         Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
//         Class.forName("easysoft.sql.jobDriver");
         conn = DriverManager.getConnection(url);
         stmt = conn.createStatement();
         rs = stmt.executeQuery(select);

         // Create the Map object.
         map = createMap(mapFilename, conn, rs);

         // Create a new DBMSToDOM object and transfer the data. Note that
         // different DOM implementations need different DocumentFactory
         // implementations.
         dbmsToDOM = new DBMSToDOM(map, new DF_Oracle2());
//         dbmsToDOM = new DBMSToDOM(map, new DF_Sun());
//         dbmsToDOM = new DBMSToDOM(map, new DF_Xerces());
         doc = dbmsToDOM.retrieveDocument(rs);

      }
      finally
      {
         // Close the result set, statement, and connection
         if (rs != null) rs.close();
         if (stmt != null) stmt.close();
         if (conn != null) conn.close();
      }

      // Write the DOM tree to a file.
      writeDocument(doc, xmlFilename);
   }

   // ***********************************************************************
   // General utility methods
   // ***********************************************************************

   static Map createMap(String mapFilename, Connection conn, ResultSet rs) throws Exception
   {
      MapFactory_MapDocument    factory;

      // Create a new map factory and create the Map.
      factory = new MapFactory_MapDocument(conn, rs, getSAXParser());
      return factory.createMap(new InputSource(getFileURL(mapFilename)));
   }

   static String getFileURL(String fileName)
   {
      File   file;

      file = new File(fileName);
      return "file:///" + file.getAbsolutePath();
   }

   // ***********************************************************************
   // Methods that use the Oracle version 2 parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************

   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Oracle parser.

      SAXParser parser;

      // Instantiate the parser and set various options
      parser = new SAXParser();
      parser.setValidationMode(true);
      return parser;
   }

   static void writeDocument(Document doc, String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Oracle parser.

      FileOutputStream xmlFile;

      // Write the DOM tree to a file.
      xmlFile = new FileOutputStream(xmlFilename);
      ((XMLDocument)doc).print((OutputStream)xmlFile);
      xmlFile.close();
   }

   // ***********************************************************************
   // Methods that use the Sun parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************

/*
   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Sun parser.

      return new com.sun.xml.parser.Parser();
   }

   static void writeDocument(Document doc, String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Sun parser.

      FileOutputStream xmlFile;

      // Write the DOM tree to a file.
      xmlFile = new FileOutputStream(xmlFilename);
      ((XmlDocument)doc).write((OutputStream)xmlFile);
      xmlFile.close();
   }
*/

   // ***********************************************************************
   // Methods that use the Xerces parser
   // Comment these methods out if you are using a different parser.
   // ***********************************************************************

/*
   static Parser getSAXParser()
   {
      // WARNING! This code is specific to the Xerces parser.

      return new SAXParser();
   }

   static void writeDocument(Document doc, String xmlFilename) throws Exception
   {
      // WARNING! This code is specific to the Xerces parser.

      FileOutputStream xmlFile;
      OutputFormat     format;
      XMLSerializer    serial;

      // Write the DOM tree to a file.
      xmlFile = new FileOutputStream(xmlFilename);
      format = new OutputFormat(doc);
      format.setIndenting(true);
      serial = new XMLSerializer((OutputStream)xmlFile, format);
      serial.asDOMSerializer().serialize(doc);
      xmlFile.close();
   }
*/
}

